@@ -1,9 +1,6 @@
-import os
 import re
-import shutil
-import urllib.request
-import urllib.error
 import zipfile
+from collections.abc import Iterable
 from dataclasses import dataclass
 from pathlib import Path
 from typing import IO
@@ -29,47 +26,10 @@ class Library:
     teaching: bool
 
 
-# aggregate full report here
-full_report: dict[str, int] = {}
-
-
 # ruff lead this alone man
 emit = print
 
 
-def do_import(fnames: str) -> None:
-    target_dir = get_app_local_dir()
-    if not os.path.isdir(target_dir):
-        os.makedirs(target_dir)
-    for f in fnames:
-        if f.startswith("http"):
-            lastpart = f.split("/")[-1]
-            emit(lastpart)
-
-            if not lastpart.endswith(".zip"):
-                lastpart = "downloaded.zip"
-            tfile = Path(target_dir) / lastpart
-            emit("Fething:", f, "->", tfile)
-            try:
-                urllib.request.urlretrieve(f, tfile)
-            except urllib.error.URLError:
-                emit(
-                    "Failed to download, ensure your VPN is operational and the target file exists!"
-                )
-        elif f.endswith(".zip"):
-            emit("Copying:", f, "->", target_dir)
-            shutil.copy(f, target_dir)
-        else:
-            emit("Not a zip file:", f)
-
-
-def get_app_local_dir() -> Path:
-    local_app_data = os.getenv("LOCALAPPDATA")
-    if local_app_data:
-        return Path(local_app_data) / "anyerplint"
-    raise Exception("Could not find LOCALAPPDATA")
-
-
 def do_check(
     libs: list[str],
     targets: list[str],
@@ -91,7 +51,7 @@ def do_check(
     # always feed the "standard library, except when unit testing (nostdlib)
 
     if not nostdlib:
-        local_app_data = get_app_local_dir()
+        local_app_data = util.get_app_local_dir()
         feed_lib(lib_vars, local_app_data)
 
     for lib in libs:
@@ -100,18 +60,8 @@ def do_check(
     has_errors = False
     all_errors: dict[str, str | dict[str, list[str]]] = {}
     for target in targets:
-        if os.path.isdir(target):
-            errs = check_dir(lib_vars, Path(target))
-        else:
-            assert target.lower().endswith(".xml")
-            try:
-                r = parse_file(target, teaching)
-            except (ElementTree.ParseError, PermissionError, UnicodeDecodeError) as e:
-                all_errors[target] = report_fatal(Path(target), e)
-                has_errors = True
-                continue
-            error_report = report(lib_vars, target, r)
-            errs = {target: error_report} if error_report else {}
+        files = util.glob_plus(target)
+        errs = check_files(lib_vars, files)
         if errs:
             all_errors.update(errs)
             has_errors = True
@@ -121,7 +71,11 @@ def do_check(
 
     if has_errors:
         emit("Errors found: >")
-        rep = sorted((k, v) for (k, v) in full_report.items())
+        error_counts = {
+            f: (sum(len(l) for l in e.values()) if not isinstance(e, str) else -1)
+            for f, e in all_errors.items()
+        }
+        rep = sorted((k, v) for (k, v) in error_counts.items())
         for line in rep:
             emit("  ", line[0], ";", line[1])
 
@@ -134,6 +88,8 @@ def write_lib(lib_vars: Library) -> None:
         f = open(fname, "wb")
         for line in lines:
             try:
+                if not line.strip().isascii():
+                    continue
                 enc = line.strip().encode()
             except UnicodeEncodeError:
                 # skip bad lines for now
@@ -143,7 +99,7 @@ def write_lib(lib_vars: Library) -> None:
 
     emit("Writing found function to:")
 
-    stdlib = get_app_local_dir()
+    stdlib = util.get_app_local_dir()
     calls_file = stdlib / "builtin_calls.txt"
     write_file(
         calls_file,
@@ -163,7 +119,7 @@ def write_lib(lib_vars: Library) -> None:
     write_file(vars_file, sorted(lib_vars.variables))
 
     loopvars_file = stdlib / "builtin_loopvars.txt"
-    write_file(loopvars_file, sorted(lib_vars.loopvariables))
+    write_file(loopvars_file, sorted({lv.lower() for lv in lib_vars.loopvariables}))
 
 
 def is_valid_call(call: str) -> bool:
@@ -187,6 +143,7 @@ def feed_lib(lib_vars: Library, libdir: Path) -> None:
             feed_set(lib_vars.variables, fobj)
         elif fname.endswith("_loopvars.txt"):
             feed_set(lib_vars.loopvariables, fobj)
+            lib_vars.loopvariables = {lv.lower() for lv in lib_vars.loopvariables}
 
     if not libdir.exists():
         return
@@ -209,28 +166,33 @@ def report_fatal(fname: Path, ex: Exception) -> str:
 
 
 def should_skip_file(fpath: Path) -> bool:
+    if fpath.suffix != ".xml":
+        return True
     with util.open_text_file(fpath) as f:
         text = f.read(10000)
         return not ("<erpConnector" in text or "<section" in text)
 
 
-def check_dir(lib_vars: Library, root: Path) -> dict[str, dict[str, list[str]]]:
-    errs = {}
-    for f in root.glob("**/*.xml"):
-        try:
-            if should_skip_file(f):
-                emit(f"{f}: SKIP nontemplate")
-                continue
-            r = parse_file(f, lib_vars.teaching)
-        except (ElementTree.ParseError, PermissionError, UnicodeDecodeError) as e:
-            # TODO: Add this to the report
-            report_fatal(f, e)
-            continue
+def check_file(lib_vars: Library, file: Path) -> str | dict[str, list[str]] | None:
+    try:
+        if should_skip_file(file):
+            emit(f"{file}: SKIP nontemplate")
+            return None
+        r = parse_file(file, lib_vars.teaching)
+    except (ElementTree.ParseError, PermissionError, UnicodeDecodeError) as e:
+        return report_fatal(file, e)
 
-        errlist = report(lib_vars, str(f), r)
+    return report(lib_vars, str(file), r)
+
+
+def check_files(
+    lib_vars: Library, files: Iterable[Path]
+) -> dict[str, str | dict[str, list[str]]]:
+    errs: dict[str, str | dict[str, list[str]]] = {}
+    for f in files:
+        errlist = check_file(lib_vars, f)
         if errlist:
             errs[str(f)] = errlist
-            full_report[str(f)] = len(errlist)
 
     return errs
 
@@ -249,24 +211,15 @@ class Parsed:
     full_content: str
 
 
-def add_linenumbers(cont: str, needles: list[str]) -> list[str]:
-    hits: list[list[int]] = [[] for i in range(len(needles))]
-    for linenum, line in enumerate(cont.splitlines()):
-        for _in, n in enumerate(needles):
-            if n in line:
-                hits[_in].append(linenum)
-
-    return [
-        n + " - line " + ", ".join(map(str, hits[idx]))
-        for (idx, n) in enumerate(needles)
-    ]
-
-
 def report(lib_vars: Library, fname: str, p: Parsed) -> dict[str, list[str]]:
     undeclared_vars = p.var_used - p.var_decl
     undeclared_vars.difference_update(lib_vars.variables)
-    unknown_loop_variables = p.loop_var_use - p.loop_var_decl
-    unknown_loop_variables.difference_update(lib_vars.loopvariables)
+    unknown_loop_variables = {
+        lv
+        for lv in p.loop_var_use
+        if lv.lower() not in p.loop_var_decl
+        and lv.lower() not in lib_vars.loopvariables
+    }
 
     if lib_vars.teaching:
         lib_vars.calls.update(p.calls)
@@ -274,12 +227,12 @@ def report(lib_vars: Library, fname: str, p: Parsed) -> dict[str, list[str]]:
         lib_vars.jumps.update(p.jumps)
         lib_vars.tags.update(p.alltags)
         lib_vars.variables.update(undeclared_vars)
-        lib_vars.loopvariables.update(unknown_loop_variables)
+        lib_vars.loopvariables.update(lv.lower() for lv in unknown_loop_variables)
 
     errors = {}
 
     if undeclared_vars:
-        errors["Unknown variables"] = add_linenumbers(
+        errors["Unknown variables"] = util.add_linenumbers(
             p.full_content,
             sorted(undeclared_vars),
         )
@@ -292,7 +245,7 @@ def report(lib_vars: Library, fname: str, p: Parsed) -> dict[str, list[str]]:
     unknown_functions = p.used_functions
     unknown_functions.difference_update(lib_vars.functions)
     if unknown_functions:
-        errors["Unknown functions"] = add_linenumbers(
+        errors["Unknown functions"] = util.add_linenumbers(
             p.full_content,
             sorted(unknown_functions),
         )
@@ -328,30 +281,31 @@ key_params: dict[str, str] = {
 
 
 def summarize_call(node: Element) -> str:
-    name = node.attrib["name"].lower()
+    name = node.attrib.get("name", "NONAME").lower()
     full = name
+    # What should we do when there's multiple params with the same name?
     params = {
-        p.attrib["name"]: (p.text or "TEXTMISSING") for p in node.iter("parameter")
+        p.attrib.get("name", "NONAME"): (p.text or "TEXTMISSING")
+        for p in node.iter("parameter")
     }
     suboperation_param_name = key_params.get(name)
     if suboperation_param_name:
         suboperation = params.get(suboperation_param_name, "UNK")
         full += "." + suboperation
 
-    return full + " - " + ",".join(sorted(params))
+    return (full + " - " + ",".join(sorted(params))).strip()
 
 
 def summarize_output(node: Element) -> str:
     type = node.attrib.get("type", "NOTYPE")
     params = {p.attrib.get("name", "NONAME") for p in node.iter("parameter")}
 
-    return "Output " + type + " - " + ",".join(sorted(params))
+    return ("Output " + type + " - " + ",".join(sorted(params))).strip()
 
 
 def summarize_tag(node: Element) -> str:
     at = " " + " ".join(sorted(node.attrib.keys())) if node.attrib else ""
-    full = "<" + node.tag + at + ">"
-    return full
+    return "<" + node.tag + at + ">"
 
 
 def brace_check(s: str) -> list[str]:
@@ -428,24 +382,6 @@ def describe_jump(n: Element) -> str:
     return desc.strip()
 
 
-def _replace_with_empty(match: re.Match[str]) -> str:
-    comment = match.group(0)
-    empty_lines = "\n" * comment.count("\n")
-    return empty_lines
-
-
-def replace_commented_xml_with_empty_lines(xml_string: str) -> str:
-    comment_pattern = "<!--(.*?)-->"
-    result = re.sub(comment_pattern, _replace_with_empty, xml_string, flags=re.DOTALL)
-    return result
-
-
-def replace_cdata_with_empty_lines(xml_string: str) -> str:
-    cdata_pattern = r"<!\[CDATA\[(.*?)\]\]>"
-    result = re.sub(cdata_pattern, _replace_with_empty, xml_string, flags=re.DOTALL)
-    return result
-
-
 def is_illegal_password(name: str, value: str) -> bool:
     if "passw" not in name.lower():
         return False
@@ -475,7 +411,7 @@ def check_single_outputresource(elem: Element, output_location: str) -> str | No
             return f"Command missing 'type' attribute in OutputResource ({output_location}, command {ind})"
         elif cmd_type in XMLWRITER_COMMAND_SINGLE:
             continue
-        elif cmd_type in XMLWRITER_COMMAND_PAIRS.keys():
+        elif cmd_type in XMLWRITER_COMMAND_PAIRS:
             cmd_stack.append(cmd_type)
         elif cmd_type in XMLWRITER_COMMAND_PAIRS.values():
             try:
@@ -504,10 +440,29 @@ def check_outputresources(root: Element) -> list[str]:
     return errors
 
 
+def find_invalid_expressions(content: str) -> list[str]:
+    expr_regex = r"{(.*?)}"
+
+    def is_invalid_expression(expr: re.Match[str]) -> bool:
+        if "," not in expr.group(1):
+            # Regular expressions can contain things like {3}
+            if expr.group(1).isnumeric():
+                return False
+            return True
+        return False
+
+    return [
+        f"Invalid expression {expr.group(0)}, line {linenum}"
+        for linenum, line in enumerate(content.splitlines(), 1)
+        for expr in re.finditer(expr_regex, line)
+        if is_invalid_expression(expr)
+    ]
+
+
 def parse_file(fname: str | Path, teaching: bool = False) -> Parsed:
     tree = ElementTree.parse(fname)
     raw_cont = util.open_text_file(fname).read()
-    comments_removed = replace_commented_xml_with_empty_lines(raw_cont)
+    comments_removed = util.replace_commented_xml_with_empty_lines(raw_cont)
 
     vardecl = {
         v.attrib.get("name", "unknown_var"): (v.text or "")
@@ -518,9 +473,11 @@ def parse_file(fname: str | Path, teaching: bool = False) -> Parsed:
         for v in tree.iter("parameter")
     }
 
+    prop_regex = r"([a-zA-Z_][a-zA-Z0-9_.]*),(\w+)"
+
     propaccess = {
         (match.group(1), match.group(2))
-        for match in re.finditer("([a-zA-Z.]+),(\\w+)", comments_removed)
+        for match in re.finditer(prop_regex, comments_removed)
     }
     varuse = {name for expr_type, name in propaccess if expr_type.lower() == "v"}
     used_functions = {
@@ -552,7 +509,7 @@ def parse_file(fname: str | Path, teaching: bool = False) -> Parsed:
     errors = []
 
     if not teaching:
-        cdata_removed = replace_cdata_with_empty_lines(comments_removed)
+        cdata_removed = util.replace_cdata_with_empty_lines(comments_removed)
         errors.extend(brace_check(cdata_removed))
 
         no_text_allowed_tags = [
@@ -584,6 +541,9 @@ def parse_file(fname: str | Path, teaching: bool = False) -> Parsed:
         outputresource_errors = check_outputresources(tree.getroot())
         errors.extend(outputresource_errors)
 
+        expr_errors = find_invalid_expressions(comments_removed)
+        errors.extend(expr_errors)
+
     return Parsed(
         var_decl=set(vardecl),
         var_used=varuse,
