@@ -1,7 +1,7 @@
 import argparse
 
 from anyerplint import __version__
-from anyerplint.cli import check, importcmd
+from anyerplint.cli import check, importcmd, initcmd, parsecmd
 
 
 def main() -> None:
@@ -24,5 +24,14 @@ def main() -> None:
     importcmd.init_parser(
         subparsers.add_parser("import", help="Import definitions from .zip file"),
     )
+
+    initcmd.init_parser(
+        subparsers.add_parser("init", help="Initialize vscode integration")
+    )
+
+    parsecmd.init_parser(
+        subparsers.add_parser("parse", help="Parse and check expressions in templates")
+    )
+
     args = main_parser.parse_args()
     args.func(args)
